/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * Provides the interface functions as defined in @see ds485-interface.h
 * Contains also some functions for ds485 sending/receiving
 */
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <pthread.h>

#include "digitalSTROM/ds485-core.h"
#include "ds485-unix.h"

#include "qp_port.h"
#include "receiver.h"
#include "digitalSTROM/ds485.h"

#ifdef DS485_API_V1
#include "ds485-qp.v1.h"
#include "digitalSTROM/ds485-const.v1.h"
#else
#include "ds485-qp.h"
#include "digitalSTROM/ds485-const.h"
#endif

#include "ds485-signals.h"
#include    "ds485-proto.h"
#include    "logger.h"
#include    "callqueue.h"
#include    "../mmi.h"

Q_DEFINE_THIS_MODULE(ds485-unix)

/************************** STATIC DATA ****************************/

Ds485LibInst_t  devInst;
static bool isOpen   = FALSE;
extern QEQueue l_ds485RxMessageQueue;
extern QEQueue l_ds485TxResponseQueue;




/************************** dS485 SYSTEM ***************************/

int ds485_receiver(unsigned char *rxBuf, unsigned int rxLen)
{
  while( rxLen > 0 ){
      rxLen--;

      switch (*rxBuf) {
      case DS485_START_BYTE: {
          static QEvent const startByteEvt = { RX_START_BYTE_SIG, 0 };
          QActive_postFIFO(AO_Receiver, &startByteEvt);
          break;
      }
      case DS485_ESCAPE_BYTE: {
          static QEvent const escapeByteEvt = { RX_ESCAPE_BYTE_SIG, 0 };
          QActive_postFIFO(AO_Receiver, &escapeByteEvt);
          break;
      }
      case DS485_RESERVED_BYTE: {
          static QEvent const reservedByteEvt = { RX_RESERVED_BYTE_SIG, 0 };
          QActive_postFIFO(AO_Receiver, &reservedByteEvt);
          break;
      }
      default: {
          OtherByteEvt* otherByteEvt = Q_NEW(OtherByteEvt, RX_OTHER_BYTE_SIG);
          otherByteEvt->byte = (uint8_t)*rxBuf;
          QActive_postFIFO(AO_Receiver, (QEvent*)otherByteEvt);
          break;
      }
      }

      rxBuf++;
  }
}


/**
 * Called by the dS485 protocol stack to immediately send a
 * frame on the wire
 */
int ds485_system_tx_raw(unsigned char *data, unsigned int length)
{
    Ds485LibInst_t* dev = &devInst;
    int result = 0;

    if (dev->tty.devfd >= 0) {
        if (write(dev->tty.devfd, data, length) != length) {
            result = -1;
        }
        tcdrain(dev->tty.devfd);
    }

    if (dev->serialLoopback) {
        ds485_receiver( data, length );

        if( dev->logFd >= 0 ) {
            if (write( dev->logFd, data, length ) != length) {
                result = -1;
            }
        }
    }

    return result;
}

/*
 * called by protocol layer to give transmission control to
 * the system send queue
 * NOTE: this is a cooperative function, the system implementation
 *       must not block the token for too long!
 * NOTE: the token is automatically forwarded by the calling protocol
 *       layer when this function returns
 */
int ds485_system_tx()
{
    Ds485LibInst_t* sysdev = &devInst;
    DS485LongMessageEvt const * txMsg;
    uint8_t* p;
    uint16_t length, n;
    static uint8_t data[1 + DS485_LONG_MESSAGE_SIZE * 2];

    txMsg = (DS485LongMessageEvt const *)DS485_GetNextTxMessage();

    if( txMsg == NULL ) Q_ERROR();  //this holds because it is exactly one message in the queue

    p = data;
    *p++ = 0xFF; /* add 0xFF to block noise */
    *p++ = DS485_START_BYTE;
    for (n = 1; n < txMsg->size; n++) {
        switch (txMsg->message[n])
        {
        case DS485_ESCAPE_BYTE:
        case DS485_START_BYTE:
        case DS485_RESERVED_BYTE:
            *p++ = DS485_ESCAPE_BYTE;
            *p++ = txMsg->message[n] & 0x7f;
            break;
        default:
            *p++ = txMsg->message[n];
            break;
        }
    }

    /* explicitly recycle the event */
    QF_gc((QEvent const *) txMsg);

    length = p - data;
    ds485_system_tx_raw(data, length);
}

#if 0
/*
 * called by dS485 protocol stack to retrieve current system time
 */
time_t ds485_system_time()
{
    struct timeval tv;
    time_t milliseconds;

    gettimeofday(&tv, NULL);
    milliseconds = tv.tv_usec / 1000;

    return tv.tv_sec * 1000 + milliseconds;
}

time_t ds485_system_time_diff_ms(time_t ts)
{
    return ds485_system_time() - ts;
}

/*
 * called by dS485 protocol stack to get a random integer number
 */
unsigned int ds485_system_random(unsigned int max)
{
    return (random() % max);
}
#endif

static void* sysRxThread(void* arg)
{
    Ds485LibInst_t *dev = (Ds485LibInst_t *) arg;
    unsigned char *p;
    int rnum, i, n;
    int numRxBytes = 0;
    int devOpenError = 0;
    uint8_t rxBuffer[512];

    /*
     * Main serial device handler
     */

    while (dev->mainActive) {

        if (strncmp("/dev", dev->devName, 4) == 0) {
           open_tty(dev->devName,&dev->tty,dev);
        }
        else {
            dev->tty.devfd = open(dev->devName, 0);
        }
        if (dev->tty.devfd < 0) {
            if (!devOpenError)
                ds485_system_msg(dev, LOG_ERR, "ERROR: cannot open %s [%d]: error = %s (%d)\n", dev->devName,
                        dev->tty.devfd, strerror(errno), errno);
            sleep(1);
            devOpenError++;
            continue;
        }
        devOpenError = 0;

        ds485_system_msg(dev, LOG_DEBUG, "Serial device open %s: %d\n", dev->devName, dev->tty.devfd);

        while (dev->mainActive && dev->tty.devfd >= 0) {

            fd_set readfs;
            struct timeval Timeout;
            int rval;

            FD_ZERO(&readfs);
            FD_SET(dev->tty.devfd, &readfs);
            Timeout.tv_usec = 0; /* microseconds */
            Timeout.tv_sec = 1; /* seconds */

            rval = select(dev->tty.devfd + 1, &readfs, NULL, NULL, &Timeout);

            /* timeout */
            if (rval == 0) {
                //printf("Serial Select Timeout [%lu]\n", ds485_system_time());
            }

            /* error */
            else if (rval < 0) {
                break;
            }

            /* data */
            else if (FD_ISSET(dev->tty.devfd, &readfs)) {

                rnum = read(dev->tty.devfd, rxBuffer, sizeof(rxBuffer));
                if (rnum == 0) {
                    if (errno == 0) {
                        ds485_system_msg(dev, LOG_DEBUG, "Serial port disconnected\n");
                        break;
                    }
                    else if (errno == 2) {
                        sleep(2);
                        break;
                    }
                }
                else if (rnum < 0) {
                    ds485_system_msg(dev, LOG_ERR, "Serial port read error: %s (%d)\n", strerror(errno), errno);
                    break;
                }
                else if (rnum > 0) {
                    ds485_receiver(rxBuffer, rnum);
                    if( dev->logFd >= 0 ) {
                        if (write( dev->logFd, rxBuffer, rnum ) != rnum) {
                            ds485_system_msg(dev, LOG_ERR, "Loopback write error\n");
                        }
                    }
                }
            }
        }

        ds485_system_msg(dev, LOG_DEBUG, "Serial port closed %d\n", dev->tty.devfd);

        close(dev->tty.devfd);
        dev->tty.devfd = -1;

    }

    return NULL;
}

extern void libqp_init(void);
extern void libqp_thread(void);

static void* sysQpThread(void* arg)
{
    ds485_system_msg(&devInst, LOG_DEBUG, "Startup QP/C libmain\n");

    libqp_thread();

    ds485_system_msg(&devInst, LOG_DEBUG, "Shutdown QP/C libmain\n");
    return NULL;
}

/********************* APPLICATION INTERFACE ***********************/

ds485Handle_t ds485_api_open(const char *device, const dsid_t *dsid)
{
    /*
     * Exclusive device
     */
    if (isOpen) {
        return NULL;
    }

    isOpen == TRUE;

    /*
     * Ressource allocation
     */
    Ds485LibInst_t* dev = &devInst;

    if (dev == NULL) {
        return NULL;
    }


    sem_init( &dev->semRx, 0, 0 );
    sem_init( &dev->semTx, 0, 0 );

    memset(dev, 0, sizeof(Ds485LibInst_t));

    dev->logFd        = -1;
    dev->tty.devfd  = -1;
    dev->devName = strdup(device);

    dev->blockingMode = 1;

    /*
     * Linux platform init
     */
    srandom(time(0));
    libqp_init();

    /*
     * start threads
     */
    dev->mainActive = 1;
    pthread_mutex_init(&dev->logMutex, NULL);

    pthread_attr_t attr;
    struct sched_param param;
    pthread_attr_init(&attr);
    pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
    param.sched_priority = sched_get_priority_min(SCHED_FIFO);
    pthread_attr_setschedparam(&attr, &param);

    if (pthread_create(&dev->qpThread, &attr, sysQpThread, dev)) {
        pthread_attr_setschedpolicy(&attr, SCHED_OTHER);
        param.sched_priority = sched_get_priority_min(SCHED_OTHER);
        pthread_attr_setschedparam(&attr, &param);
        if (pthread_create(&dev->qpThread, &attr, sysQpThread, dev)) {
        }
    }

    if (pthread_create(&dev->rxThread, NULL, sysRxThread, dev)) {
    }

    DatamodelConfigEvt * evt = Q_NEW(DatamodelConfigEvt, CONFIG_INIT_SIG);
    memcpy( &evt->dsmid, dsid, sizeof(dsid_t) );
    QActive_postFIFO(AO_DS485, (QEvent *)evt);

    ccCreate( &dev->reqBusMembers, dev->reqBusMembersSto, Q_DIM(dev->reqBusMembersSto) );
    ccCreate( &dev->reqBusState, dev->reqBusStateSto, Q_DIM(dev->reqBusStateSto) );

    return dev;
}

int ds485_api_close( ds485Handle_t handle )
{
    Ds485LibInst_t* dev = handle;

    if( !isOpen )   return  -1;
    /*
     * Stop threads & timer
     */
    dev->mainActive = 0;

    QF_stop();

    ccFree( &dev->reqBusMembers );
    ccFree( &dev->reqBusState );

    /*
     * Cleanup rx/tx threads
     */
    pthread_cancel(dev->rxThread);

    sem_destroy( &dev->semRx );
    sem_destroy( &dev->semTx );

    /*
     * Restore serial terminal settings
     */
    close_tty(&dev->tty,dev);

    /*
     * Cleanup memory ressources
     */
    free(dev->devName);
    isOpen = FALSE;

    return 0;
}

int ds485_api_send_data( ds485Handle_t handle, ds485_container_t const *packet )
{
    Ds485LibInst_t* dev = handle;

    if( dev == NULL ) return    -1;

#ifdef DS485_API_V1
    if( packet->length > DS485_SHORT_PAYLOAD_SIZE ) return    -1;
    DS485DataMessageEvt * evt = DS485_CreateDataMessage( &packet->destinationId, DS485_TYPE_REQ, packet->length);
    memcpy( &evt->message[DS485_DATA_OFFSET], packet->data, packet->length);
#else
    if( packet->length > DS485_LONG_PAYLOAD_SIZE ) return    -1;
    DS485LongMessageEvt * evt = DS485_CreateLongDataMessage( &packet->destinationId, DS485_TYPE_REQ, packet->length + 1 );
    memcpy( &evt->message[DS485_DATA_OFFSET], packet->raw_data, packet->length + 1);
#endif

    QActive_postFIFO(AO_DS485, (QEvent *) evt);

    if( dev->blockingMode ){
        return  waitTxResponse();
    } else {
        return  0;
    }
}

int ds485_api_recv_data( ds485Handle_t handle, ds485_container_t* packet )
{
    Ds485LibInst_t* dev = handle;

    if( !dev->blockingMode ){
        int val;
        sem_getvalue( &dev->semRx, &val );
        if( val <= 0 ){
            return  -1;
        }
    }

    sem_wait( &dev->semRx );

    DS485PacketEvt const * msg;
    msg = (DS485PacketEvt const *)QEQueue_get(&l_ds485RxMessageQueue);
    memcpy( packet, &msg->packet, sizeof(*packet) );
    QF_gc((QEvent const *) msg);

    return    0;
}

int ds485_api_ioctl( ds485Handle_t handle, unsigned int cmd, void* data, unsigned int length )
{
    Ds485LibInst_t* dev = handle;

    int rval = 0;

    switch (cmd)
    {
    case IOC_SET_MSG_CB:
        dev->callback_msg_handler = (ds485_api_callback_msg_t) data;
        break;

    case IOC_SET_LOGGER_FD:
        dev->logFd    = *(int*) data;
        break;

    case IOC_SET_BUS_CHANGE_CB:
        dev->busChange = (ds485_api_callback_bus_change_t) data;
        break;

    case IOC_SET_RECEIVED_CB:
        dev->received = (ds485_api_callback_received_t) data;
        break;

    case IOC_SET_SENDER_CB:
        dev->sender = (ds485_api_callback_sender_t) data;
        break;

    case IOC_SET_BUS_STATE_CB:
        dev->state = (ds485_api_callback_state_t) data;
        break;

    case IOC_LOOPBACK_MODE:
        dev->serialLoopback = *(unsigned int *) data;
        break;

    case IOC_BLOCKING_MODE:
        dev->blockingMode = *(unsigned int *) data;
        break;
    }

    return rval;
}

int ds485_api_get_bus_members( ds485Handle_t handle, dsid_t *members, int maxNumberOfMembers )
{
    Ds485LibInst_t* dev = handle;

    int nr  = 0;

    QActive_postFIFO( AO_DS485, ccPostRequest( &dev->reqBusMembers, DS485_REQ_BUS_MEMBERS_SIG ) );
    const DS485_DsidResponseEvt   *res    = (const DS485_DsidResponseEvt*)ccWaitResponse( &dev->reqBusMembers );

    int i;
    for( i = 0; i < DS485_MAX_SLAVE_ADDRESS+1; i++ ){
        if( (nr < maxNumberOfMembers) && (!IsNullId(res->dsid[i])) ){
            memcpy( members, &res->dsid[i], sizeof(dsid_t) );
            members++;
            nr++;
        }
    }
    QF_gc((QEvent const *) res);
    return  nr;
}

bus_state_t ds485_api_get_bus_state( ds485Handle_t handle )
{
    Ds485LibInst_t* dev = handle;

    QActive_postFIFO( AO_MMI, ccPostRequest( &dev->reqBusState, MMI_REQ_BUS_STATE_SIG ) );
    const MMI_BusStateEvt   *res    = (const MMI_BusStateEvt*)ccWaitResponse( &dev->reqBusState );

    bus_state_t    state   = res->state;

    QF_gc((QEvent const *) res);
    return  state;
}
